@@ -8,7 +8,6 @@ use File::Spec;
 use Sys::Hostname;
 use DBI;
 use Helios::ObjectDriver::DBI;
-require XML::Simple;
 
 use Helios::Error;
 use Helios::Job;
@@ -16,8 +15,108 @@ use Helios::Config;
 use Helios::ConfigParam;
 use Helios::LogEntry;
 use Helios::LogEntry::Levels qw(:all);
+# [LH] [2013-10-04]: Using Helios::JobType instead of TheSchwartz::FuncMap now.
+use Helios::JobType;
+use Helios::Error::JobTypeError;
+
+our $VERSION = '2.61_4051';
+
+# FILE CHANGE HISTORY:
+# [2011-12-07]: Updated to support new Helios::Logger API.  Added 
+# %INIT_LOG_CLASSES global.  Completely rewrote logMsg() method.  Updated 
+# logMsg() documentation.
+# [2011-12-07]: Updated copyright info.
+# [2011-12-07]: Removed parseArgXML() method (redundant).
+# [2011-12-28]: Removed unnecessary 'use XML::Parser' line.
+# [2011-12-28]: Replaced metajob running code in work() with new runMetajob()
+# method.
+# [2011-12-28]: work(): changed so CACHED_CONFIG and 
+# CACHED_CONFIG_RETRIEVAL_COUNT only print to STDOUT if debug() is set.
+# [2011-12-28]: Updated copyright info.
+# [2012-01-01]: Renamed runMetajob() method to metarun().
+# [2012-01-01]: work(): replaced old code calling the service class's run() 
+# method.  New code: 1) calls run() or metarun() as appropriate, 2) ignores 
+# value returned by run() and metarun() unless DOWNSHIFT_ON_NONZERO_RUN 
+# parameter is set, 3) is wrapped in an eval {} to catch uncaught exceptions
+# a service class's run() might throw, forcing the job to failure.  Updated
+# work() documenation for new functionality.
+# [2012-01-01]: Updated copyright info for new year.
+# [2012-01-04]: Fixed max_retries() and retry_delay() so they actually pay 
+# attention to MaxRetries() and RetryInterval().  In the original code they 
+# didn't, and MaxRetries() and RetryInterval() did not work as documented.
+# [2012-01-08]: work(): explicitly return 0 to the calling routine. 
+# [2012-03-27]: Reorganized use module lines.  Removed unnecessary TheSchwartz &
+# TheSchwartz::Job lines.
+# [2012-03-27]: work(): added debugging code for new driver and logger code.
+# [2012-03-27]: work(): changed try {} to eval {}.
+# [2012-03-27]: prep(): Replaced old prep() method with new version that 
+# starts new logger and config initialization.
+# [2012-03-27]: jobsWaiting(): changed quote operator for query from heredoc 
+# to qq{}
+# [2012-03-27]: added new setDriver() and initDriver() methods.  Replaced 
+# getDriver() method with new one that uses setDriver() and initDriver().
+# [2012-03-27]: added initLoggers() method to handle logger module 
+# initialization.
+# [2012-04-25]: added deferredJob() method.
+# [2012-05-20]: work: removed driver and logger debugging code.  Removed 
+# comment about removing a debug message before release (it is useful to leave 
+# that debugging message in).
+# [2012-05-20]: dbConnect(): removed old commented-out code.
+# [LH] [2012-07-11]: Switched use line for Data::ObjectDriver::Driver::DBI to 
+# load Helios::ObjectDriver::DBI to start integration of database connection 
+# caching.
+# [LH] [2012-07-15]: Changed prep() to use new Helios::Config class.  Removed 
+# 'use Config::IniFiles' because with Helios::Config it's redundant.
+# [LH] [2012-07-15]: replaced most of dbConnect() code to implement fork-safe 
+# database connection creation and sharing.
+# [LH] [2012-07-15]: replaced most of jobWaiting() code for simplicity and to 
+# replace try{} with eval {}.
+# [LH] [2012-07-15]: replaced most of getFuncidFromDb() code to change try{} 
+# to eval{} and eliminate indirect object notation.
+# [LH] [2012-07-16]: getFuncidFromDb(): fixed identation of new code.
+# [LH] [2012-07-16]: updated copyright notices (added Logical Helion, LLC to 
+# main COPYRIGHT section).
+# [LH] [2012-08-04]: removed 'use Error' line as all of the try {} blocks have
+# been replaced with eval {}.
+# [LH] [2012-08-04]: replaced getConfigFromIni() and getConfigFromDb() with 
+# versions that use the new Helios::Config API.  Changed POD for both to note
+# the methods are deprecated.
+# [LH] [2012-08-04]: added new initConfig() method to manage Helios::Config 
+# module initialization.
+# [LH] [2012-08-04]: added blank default ConfigClass() method.
+# [LH] [2012-08-04]: dbConnect(): updated to better handle "options" directives
+# and improve connection code.  Updated dbConnect() POD.
+# [LH] [2012-08-04]: Reformatted copyright notices for clarity.
+# [LH] [2012-08-07]: further changes to getConfigFromIni() and 
+# getConfigFromDb() to work with Helios::Config API.
+# [LH] [2012-09-05]: removed old commented out code from getConfigFromIni(), 
+# getConfigFromDb(), getFuncidFromDb(), dbConnect().
+# [LH] [2012-09-05]: Added to POD entry for getFuncidFromDb().
+# [LH] [2012-11-06]: Added _require_module() method to safely load modules at 
+# runtime.
+# [LH] [2012-11-06]: removed old commented out 'use' lines for 
+# Config::IniFiles, Data::ObjectDriver::Driver::DBI, Error.
+# [LH] [2012-11-06]: corrected grammar in work() documentation.
+# [LH] [2012-11-06]: removed old commented out code from prep().
+# [LH] [2012-11-06]: removed old commented out code from getDriver().
+# [LH] [2012-11-06]: Added ConfigClass() and initConfig() POD.
+# [LH] [2013-08-11]: Added code to work() to catch and handle job 
+# initialization errors.  [RT79690]
+# [LH] [2013-08-19]: Removed old commented out code and clarified comments on 
+# job initialization error handling.
+# [LH] [2013-10-04]: Added code to start conversion to new Helios class 
+# structure and support virtual jobtypes feature.  New new() constructor 
+# initializes attribute hashref values and can only be called as a class 
+# method.  Added set/get/addJobType(), set/getAltJobTypes(), 
+# set/getAltJobtypeids(), addAltJobtypeid(), lookupJobtypeid(), 
+# lookupAltJobtypeids().  Switched all code that used TheSchwartz::FuncMap to 
+# use Helios::JobType.  Replaced set/getFuncid() with new version that mirrors 
+# set/getJobType() (set/getFuncid() will be deprecated on final release).  
+# Replaced jobsWaiting() with new version that uses Helios::JobType and scans 
+# for all jobtypes (primary and alternates) if alternate jobtypes are set.
+# [LH] [2013-10-04] Removed 'require XML::Simple' line because Helios::Service 
+# has not used that in a long time.
 
-our $VERSION = '2.52_4150';
 
 =head1 NAME
 
@@ -95,7 +194,29 @@ sub retry_delay { $_[0]->RetryInterval(); }
 sub work {
 	my $class = shift;
 	my $schwartz_job = shift;
-	my $job = $class->JobClass() ? $class->JobClass()->new($schwartz_job) : Helios::Job->new($schwartz_job);
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+	# 2013-08-11: Rewritten job initialization code to catch job init errors, including [RT79690].
+	my $job;
+	my $job_init_error;
+	eval {
+		# turn the schwartz job we were given into: 
+		# a custom job object defined by the app class,
+		# or a basic Helios::Job object if the app didn't specify anything special
+		if ( $class->JobClass() ) {
+			# instantiate a custom job object
+			$job = $class->JobClass()->new($schwartz_job);
+		} else {
+			# nothing fancy, just a normal Helios::Job object
+			$job = Helios::Job->new($schwartz_job);
+		}
+		1;
+	} or do {
+		# uhoh, there was a problem turning the schwartz job into a Helios job
+		# note that, and when the worker is fully prepped, 
+		# we'll take care of the problem
+		$job_init_error = "$@";
+	};
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
 	$WORKER_START_TIME = $WORKER_START_TIME ? $WORKER_START_TIME : time();     # for WORKER_MAX_TTL 
 	my $return_code;
 	my $args;
@@ -132,6 +253,20 @@ sub work {
                 $CACHED_CONFIG_RETRIEVAL_COUNT = 1;     # "prime the pump"
             }	    
         }
+
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+		# 2013-08-11: Rewritten job initialization code to catch job init errors, including [RT79690].
+		# if a job initialization error occurred above,
+		# we want to log the error and then exit the worker process
+		# trying to further job setup and/or run the job is ill-advised,
+		# and if we have to exit the process so TheSchwartz doesn't force the job to failure.
+		# (but we have to wait and do it here so we can properly log the error)
+		if ( defined($job_init_error) ) {
+			if ($self->debug) { print "JOB INITIALIZATION ERROR: ".$job_init_error."\n"; }
+			$self->logMsg(LOG_CRIT, "JOB INITIALIZATION ERROR: $job_init_error");
+			exit(1);
+		}
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
 	    	    
 		$job->debug( $self->debug );
 		$job->setConfig($self->getConfig());
@@ -219,6 +354,7 @@ sub work {
 	return 0;
 }
 
+# BEGIN CODE Copyright (C) 2011-2012 by Andrew Johnson.
 
 =head2 metarun($job)
 
@@ -240,13 +376,6 @@ have thousands (or tens of thousands, or more) of jobs to run, especially if
 you are running your service in OVERDRIVE mode, you should use metajobs to 
 greatly increase system throughput.
 
-=head3 COPYRIGHT
-
-This method is Copyright (C) 2011-2012 by Andrew Johnson.
-
-See the COPYRIGHT AND LICENSE section elsewhere in this document for specific
-copyright and license terms.
-
 =cut
 
 sub metarun {
@@ -280,6 +409,7 @@ sub metarun {
 		}
 	};
 }
+# END CODE Copyright (C) 2011-2012 by Andrew Johnson.
 
 
 =head1 ACCESSOR METHODS
@@ -316,8 +446,11 @@ sub getJobType { return $_[0]->{jobType}; }
 sub setConfig { $_[0]->{config} = $_[1]; }
 sub getConfig { return $_[0]->{config}; }
 
-sub setFuncid { $_[0]->{funcid} = $_[1]; }
-sub getFuncid { return $_[0]->{funcid}; }
+# [LH] [2013-10-04]: Virtual jobtypes.  Changed set/getFuncid() for 
+# compatibility with set/getJobtypeid().  Set/getFuncid() is DEPRECATED;
+# retained for now for backward compatibility with Helios 2.6x and earlier.
+sub setFuncid { $_[0]->{jobtypeid} = $_[1]; }
+sub getFuncid { return $_[0]->{jobtypeid}; }
 
 sub setIniFile { $_[0]->{inifile} = $_[1]; }
 sub getIniFile { return $_[0]->{inifile}; }
@@ -350,7 +483,7 @@ It does set the job type for the object (available via the getJobType() method).
 
 =cut
 
-sub new {
+sub new_old {
 	my $caller = shift;
 	my $class = ref($caller) || $caller;
 #	my $self = $class->SUPER::new(@_);
@@ -492,6 +625,7 @@ sub getConfigFromIni {
 	my $inifile = shift;
 
 # BEGIN CODE Copyright (C) 2012 by Logical Helion, LLC.
+# getConfigFromIni() is no longer necessary.
 	
 	unless ($INIT_CONFIG_CLASS) {
 		if ( defined($inifile) ) { $self->setIniFile($inifile); }
@@ -529,6 +663,7 @@ sub getConfigFromDb {
 	my $params = $self->getConfig();
 
 # BEGIN CODE Copyright (C) 2012 by Logical Helion, LLC.
+# getConfigFromDb() method is no longer necessary.
 
 	unless ($INIT_CONFIG_CLASS) {
 		$INIT_CONFIG_CLASS = $self->initConfig();
@@ -591,14 +726,13 @@ waiting.  Only meant for use with the helios.pl service daemon.
 
 =cut
 
-sub jobsWaiting {
+sub jobsWaiting_old {
 	my $self = shift;
 	my $params = $self->getConfig();
 	my $jobType = $self->getJobType();
 
 
 # BEGIN CODE Copyright (C) 2012 by Logical Helion, LLC.
-
 	my $jobsWaiting;
 	my $funcid = $self->getFuncid();
 	eval {
@@ -628,6 +762,8 @@ sub jobsWaiting {
 }
 
 
+# BEGIN CODE Copyright (C) 2012 by Andrew Johnson.
+
 =head2 initDriver()
 
 Creates a Data::ObjectDriver object connected to the Helios database and 
@@ -640,7 +776,6 @@ connections to the Helios database.
 
 =cut
 
-# BEGIN CODE Copyright (C) 2012 by Andrew Johnson.
 sub initDriver {
 	my $self = shift;
 	my $config = $self->getConfig();
@@ -886,7 +1021,6 @@ as a Helios::Error::LoggingError exception.
 =cut
 
 # BEGIN CODE Copyright (C) 2009-12 by Andrew Johnson.
-
 sub logMsg {
 	my $self = shift;
 	my @args = @_;
@@ -1012,6 +1146,7 @@ error.
 =cut
 
 # BEGIN CODE Copyright (C) 2012 by Andrew Johnson.
+
 sub initLoggers {
 	my $self = shift;
 	my $config = $self->getConfig();
@@ -1269,6 +1404,165 @@ sub _require_module {
 # END CODE Copyright (C) 2012 by Logical Helion, LLC.
 
 
+# BEGIN CODE Copyright (C) 2013 by Logical Helion, LLC.
+
+# [LH] [2013-10-04] Virtual jobtype support code!  All this code adds
+# support to the class for "alternate jobtypes" and makes sure jobsWaiting()
+# takes all jobtypes into account when jobs are scanned for in the JOB table.
+
+#[] this code should be distributed to their respective places in the class
+# before final release.  During development and testing, they can stay here 
+# for now since they're all here to support virtual jobtypes.
+
+# [LH] [2013-10-04] Virtual jobtype code.
+sub setJobtypeid { $_[0]->{jobtypeid} = $_[1]; }
+sub getJobtypeid { return $_[0]->{jobtypeid}; }
+
+# [LH] [2013-10-04] Virtual jobtype code.
+sub setAltJobTypes {
+	my $self = shift;
+	$self->{altJobTypes} = [@_];
+}
+sub getAltJobTypes {
+	if ( defined $_[0]->{altJobTypes} ) {
+		return @{ $_[0]->{altJobTypes} };
+	} else {
+		return undef;
+	}
+}
+sub addAltJobType {
+	push(@{ $_[0]->{altJobTypes} }, $_[1]);
+}
+
+# [LH] [2013-10-04] Virtual jobtype code.
+sub setAltJobtypeids {
+	my $self = shift;
+	$self->{altJobtypeids} = [@_];
+}
+sub getAltJobtypeids {
+	if ( defined $_[0]->{altJobtypeids} ) {
+		return @{ $_[0]->{altJobtypeids} };
+	} else {
+		return undef;
+	}
+}
+sub addAltJobtypeid {
+	push(@{ $_[0]->{altJobtypeids} }, $_[1]);
+}
+
+# [LH] [2013-10-04] Virtual jobtype code.
+=head2 lookupAltJobtypeids(@jobtypenames)
+
+=cut
+
+sub lookupAltJobtypeids {
+	my $self = shift;
+	my @jobtypes = @_ || $self->getAltJobTypes();
+	my $config = $self->getConfig();
+	my @ids;
+	
+	for (@jobtypes) {
+		my $jtid = $self->lookupJobtypeid($_);
+		unless ($jtid) { Helios::Error::JobTypeError->throw("lookupAltJobtypeids(): $_ cannot be found in collective database."); }
+		push(@ids, $jtid);
+		$self->addAltJobtypeid($jtid);
+	}
+	return @ids;
+}
+
+=head2 lookupJobtypeid($jobtypename)
+
+=cut
+
+sub lookupJobtypeid {
+	my $self = shift;
+	my $jt = shift;
+
+	my $jobtype = Helios::JobType->lookup(name => $jt, config => $self->getConfig());
+	if ($jobtype) {
+		return $jobtype->getJobtypeid();
+	} else {
+		return undef;
+	}
+}
+
+
+# [LH] [2013-10-04] jobsWaiting() replaced with new version for virtual 
+# jobtypes.    
+sub jobsWaiting {
+	my $self = shift;
+	my $num_of_jobs = 0;
+	my $primary_jobtypeid = $self->getJobtypeid();
+	my @alt_jobtypeids;
+	my $sth;
+	eval {
+		my $dbh = $self->dbConnect();
+		unless ( defined($primary_jobtypeid) ) {
+			$primary_jobtypeid = $self->lookupJobtypeid($self->getJobType);
+			$self->setJobtypeid($primary_jobtypeid);
+		}
+		if ( $self->getAltJobTypes() ) {
+			if ( $self->getAltJobtypeids() ) {
+				@alt_jobtypeids = $self->getAltJobtypeids();
+			} else {
+				@alt_jobtypeids = $self->lookupAltJobtypeids();
+			}
+		}
+		
+		if (@alt_jobtypeids) {
+			my @plhrs = ('?');	# one for the primary
+			for (@alt_jobtypeids) { push(@plhrs,'?'); }
+			my $plhrs_str = join(',' => @plhrs);
+			
+			$sth = $dbh->prepare_cached("SELECT COUNT(*) FROM job WHERE funcid IN($plhrs_str) AND (run_after < ?) AND (grabbed_until < ?)");
+			$sth->execute($primary_jobtypeid, @alt_jobtypeids, time(), time());
+		} else {
+			$sth = $dbh->prepare_cached('SELECT COUNT(*) FROM job WHERE funcid = ? AND (run_after < ?) AND (grabbed_until < ?)');
+			$sth->execute($primary_jobtypeid, time(), time());
+		}
+		my $r = $sth->fetchrow_arrayref();
+		$sth->finish();
+		$num_of_jobs = $r->[0];
+		
+		1;
+	} or do {
+		my $E = $@;
+		Helios::Error::DatabaseError->throw("$E");
+	};
+	
+	return $num_of_jobs;
+}
+
+
+# [LH] [2013-10-04] New constructor initializes attributes in the underlying
+# object structure and can only be called as a class method.  
+sub new {
+	my $cl = shift;
+	my $self = {
+		'jobType'       => undef,
+		'altJobTypes'   => undef,
+		'jobtypeid'     => undef,
+		'altJobtypeids' => undef,
+		'hostname'      => undef,
+		'inifile'       => undef,
+		'job'           => undef,
+		
+		'config' => undef,
+		'debug'  => undef,
+		'errstr' => undef,
+	};
+	bless $self, $cl;
+
+	# init fields
+	my $jobtype = $cl;
+	$self->setJobType($jobtype);
+
+	return $self;
+}
+
+
+# END CODE Copyright (C) 2013 by Logical Helion, LLC.
+
 
 
 1;
@@ -1278,7 +1572,7 @@ __END__
 =head1 SEE ALSO
 
 L<Helios>, L<helios.pl>, L<Helios::Job>, L<Helios::Error>, L<Helios::Config>, 
-L<TheSchwartz>
+L<Helios::JobType>
 
 =head1 AUTHOR
 
@@ -1295,7 +1589,7 @@ Portions of this software, where noted, are
 Copyright (C) 2011-2012 by Andrew Johnson.
 
 Portions of this software, where noted, are
-Copyright (C) 2012 by Logical Helion, LLC.
+Copyright (C) 2012-3 by Logical Helion, LLC.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself, either Perl version 5.8.0 or,
